#include "hlserver.h"
#include "hxd_rcv.h"
#include "chatproto.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include "xmalloc.h"
#include "protocols.h"

extern htlc_t *htlc_list;
static u_int32_t notice_linec;

static void cp_snd_linef (htlc_t *htlc, const char *fmt, ...);
static void cp_snd_notice (htlc_t *htlc, const char *fmt, ...);

static struct protocol *irc_proto = 0;

/* this function remove any non-standard character in nicknames
 * and puts a ~ in front of hotline users, making their nick both unique to irc
 * and easily identifiable */
static char *
satanise_nick (htlc_t *htlc)
{
	char *final;
/*go up to first strange char, add UID */

}

static char *
crlf_to_cr ( char *str )
{
	u_int8_t i;
	char *final;
	for (i=0;str[i]; i++)
		if (str[i] != '\n')
			strcpy(final[i], str[i]);
}


static char *
read_file (int fd, size_t max, size_t *lenp)
{
	char *__news_buf = 0;
	size_t __news_len = 0;
#define BLOCKSIZE	4096
	char *buf = __news_buf;
	size_t len = __news_len;
	size_t pos = 0;
	size_t rn;
	ssize_t r;

	for (;;) {
		if (pos+BLOCKSIZE > len) {
			len += BLOCKSIZE;
			buf = xrealloc(buf, len);
		}
		rn = max > BLOCKSIZE ? BLOCKSIZE : max;
		r = read(fd, buf+pos, rn);
		if (r <= 0) {
			if (errno == EINTR || errno == EAGAIN)
				continue;
			break;
		}
		pos += r;
		max -= r;
		if (r != (ssize_t)rn || !max)
			break;
	}
	if (lenp)
		*lenp = pos;
	__news_buf = buf;
	__news_len = len;

	return buf;
}


static void
do_command (htlc_t *htlc, char *linep)
{
	u_int8_t *p, *cmd, *arg, *arg2;
	char *agbuf;
	u_int32_t fd, len;
	struct hl_userlist_hdr *uh;
	struct htlc_chat *chat;
	char *nick_list;
	struct htlc_conn *htlcp;
	
	u_int8_t ircops=0, users=0, invisible=0;

	/* update idle time for irc */
	gettimeofday(&htlc->idle_tv, 0);
	 
	cmd = linep;
	for (p = linep; *p; p++) {
		if (isspace(*p))
			break;
	}
	if (!*p)
		return;
	*p = 0;
	p++;
	while (isspace(*p))
		p++;
	arg = p;
	if (!*p)
		return;

	switch (cmd[0]) {
		case 'J':
			if (!strncmp(cmd, "JOIN", 4)) {
				cp_snd_linef(htlc, ":%s!%s@localhost JOIN :%s",
				htlc->name, htlc->userid, arg);
				cp_snd_linef(htlc, ":localhost MODE %s +nt", arg);
				//cp_snd_notice(htlc, "%s = %s :@%s",
				//		htlc->name, arg, htlc->name);
				cp_snd_notice(htlc, "JOIN: %s", arg);
				chat = chat_lookup_ref(0);
				cp_snd_notice(htlc, "%s", chat->subject);
				for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next)
					if (!htlcp->flags.visible)
						sprintf(nick_list, "%s", htlcp->name);
				cp_snd_notice(htlc, "= %s %s :End of /NAMES list.",
				arg, nick_list, arg);
			}
			break;
		case 'N':
			if (!strncmp(cmd, "NICK", 4)) {
				strlcpy(htlc->name, arg, sizeof(htlc->name));
				htlc->access.read_chat = 1;
				htlc->access.send_chat = 1;
				htlc->access.send_msgs = 1;
				htlc->uid = uid_assign(htlc);
				user_loginupdate(htlc);
				
				if (!htlc->access_extra.can_login)
					return;

				htlc->access_extra.can_login = 0;
				
				/*starts login emulation in IRC*/

				for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
					if(htlcp->access.disconnect_users)
						ircops++;
					if(!htlcp->flags.visible)
						invisible++;
					users++;
				}

				cp_snd_linef(htlc,
"NOTICE AUTH :*** Looking up your hostname...\r\n"
"NOTICE AUTH :*** Checking Ident\r\n");
				if(!htlc->userid) {
					strlcpy(htlc->userid, "~", 2);
					cp_snd_linef(htlc, "NOTICE AUTH :*** No Ident response");
				}

				cp_snd_notice(htlc, "Welcome to %s\r\n", hxd_cfg.tracker.name);
				cp_snd_notice(htlc, "This is an experimental mhxd server");
				cp_snd_linef(htlc, "NOTICE %s :*** To join Public Hotline Chat, type /join #public", htlc->name);
				cp_snd_notice(htlc, "WALLCHOPS KNOCK EXCEPTS INVEX MODES=4 MAXCHANNELS=65335 MAXBANS=65335 MAXTARGETS=4 NICKLEN=32 TOPICLEN=300 KICKLEN=300 :are maybe supported by this server");
				cp_snd_notice(htlc, "CHANTYPES=# PREFIX=(ohv)@%%+ CHANMODES=eIb,k,l,imnpstZ NETWORK=%s CASEMAPPING=rfc1459 CALLERID :are maybe supported by this server", "laym3n3t");

				cp_snd_notice(htlc,
"There are %u users and %u invisible on 1 servers\r\n"
"%u :IRC Operators online\r\n"
"<unknown> :channels formed\r\n"
"I have %u clients and 1 servers\r\n"
"Current local  users: %u  Max: 65335\r\n"
"Current global users: %u  Max: 65335",
users, invisible, ircops, users, users, users);

				/* We send agreement as message of the day */
				cp_snd_notice(htlc, "- Message of the Day -");
				cp_snd_notice(htlc, "- By entering this server, you agree the following statements -");
				if ((fd = SYS_open(hxd_cfg.paths.agreement, O_RDONLY, 0)) > 0)
					agbuf = read_file(fd, 0xffff, &len);
				else
					agbuf = 0;
				close(fd);
				if (agbuf)
					cp_snd_notice(htlc, agbuf);
				cp_snd_notice(htlc, "End of /MOTD command.");
				cp_snd_linef(htlc, ":%s MODE %s :+iph", arg, arg);
				cp_snd_linef(htlc, "PING :localhost");
			}
			break;
		
		case 'P':
			if (!strncmp(cmd, "PRIVMSG", 7)) {
				arg2 = arg;
				while (*arg2 && !isspace(*arg2))
					arg2++;
				*arg2 = 0;
				arg2++;
				if (*arg == '#') {
					hxd_rcv_chat(htlc, 0, 0, 0, arg2, strlen(arg2));
				} else {
					htlc_t *htlcp;
					for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next)
						if (!strcmp(htlcp->name, arg))
							break;
					
				if (!htlcp)
					return;
				hxd_rcv_msg(htlc, htlcp, arg2, strlen(arg2));
				}
	
			} else if (!strncmp(cmd, "PING", 4)) {
				cp_snd_linef(htlc, ":%s PONG localhost :%s", hxd_cfg.irc.hostname, htlc->name);
			}
			break;
			
		case 'Q':
			if (!strncmp(cmd, "QUIT", 4)) {
				snd_user_part(htlc);
				htlc_close(htlc);
			}
			break;
	}
}

static void
rcv_chat (htlc_t *htlc)
{
	struct qbuf *in = &htlc->in;
	u_int8_t *p, *linep;
	u_int32_t pos;

	linep = in->buf;
	for (p = in->buf+1; p < (in->buf+in->pos); p++) {
#if 0
		/* RFC/irssi behavior */
		if (*(p-1) == '\r' && *p == '\n') {
			*(p-1) = 0;
#else
		/* Handle mirc/bitchx as well */
		if (*p == '\n') {
			if (*(p-1) == '\r')
				*(p-1) = 0;
			else
				*p = 0;
#endif
			do_command(htlc, linep);
			linep = p+1;
		}
	}
	pos = p - linep;
	qbuf_set(in, pos, 0);
}

static int
rcv_magic (htlc_t *htlc)
{
	struct qbuf *in = &htlc->in;

	if (in->pos >= 4
	    && (   !memcmp(in->buf, "NICK", 4)
		|| !memcmp(in->buf, "USER", 4)
		|| !memcmp(in->buf, "PASS", 4))) {
		htlc->proto = irc_proto;
		htlc->flags.is_irc = 1;
		htlc->rcv = rcv_chat;
		if (in->pos > 4)
			rcv_chat(htlc);
		return 1;
	}
	if (in->pos >= 4 || (in->buf[0] != 'N' && in->buf[0] != 'P' && in->buf[0] != 'U'))
		return -1;

	return 0;
}

static void
cp_snd_line (htlc_t *htlc, u_int8_t *buf, u_int16_t len)
{
	struct qbuf *out = &htlc->out;
	qbuf_add(out, buf, len);
	qbuf_add(out, "\r\n", 2);
	if (htlc->wfd)
		hxd_fd_set(htlc->wfd, FDW);
	else
		hxd_fd_set(htlc->fd, FDW);
}

static void
cp_snd_linef (htlc_t *htlc, const char *fmt, ...)
{
	char buf[4096];
	int len;
	va_list ap;

	va_start(ap, fmt);
	len = vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	cp_snd_line(htlc, buf, len);
}

static void
cp_snd_notice (htlc_t *htlc, const char *fmt, ...)
{
	char buf [4096];
	int len;
	char line[3];
	va_list ap;
	
	notice_linec++;

	if(notice_linec < 10)
		snprintf(line, 4, "00%u", notice_linec);
	else if(notice_linec < 100)
		snprintf(line, 4, "0%u", notice_linec);
	else
		snprintf(line, 4, "%u", notice_linec);
	
	len = snprintf(buf, sizeof(buf), ":%s %s %s: ", hxd_cfg.irc.hostname, line, htlc->name); 
	va_start(ap, fmt);
	len += vsnprintf(buf+len, sizeof(buf), fmt, ap);
	va_end(ap);
	
	cp_snd_linef(htlc, buf, len);
}

static void
snd_msg (htlc_t *htlc, htlc_t *htlcp, u_int8_t *buf, u_int16_t len)
{
	char host[255];
	char abuf[HOSTLEN+1];

	inaddr2str(abuf, &htlcp->sockaddr);
	if (!htlc->host)
		snprintf(host, 255, abuf);
	else
		snprintf(host, 255, htlc->host);
	
	cp_snd_linef(htlcp, ":%s!%s@%s PRIVMSG %s :%.*s", htlc->name, htlc->userid, host, htlcp->name, len, buf);
}

static void
snd_chat (htlc_t *htlc, struct htlc_chat *chat, htlc_t *from, u_int8_t *buf, u_int16_t len)
{
	u_int8_t *name, *p;

	
	/* in IRC, we don't send back the chat to the sender */
	if (from == htlc)
		return;
	
	/* XXX put ip if host is not resolved */
	p = buf;
	while (isspace(*p))
		p++;
	name = p;
	for (; p < buf+len;p++) {
		if (*p == ':')
			break;
	}
	*p = 0;
	buf[len] = 0;
	cp_snd_linef(htlc, ":%s!%s@%s PRIVMSG %s %s", name, htlc->userid, htlc->host, "#public", p+2);
	*p = ':';
}

static void
snd_user_part (struct htlc_conn *to, struct htlc_conn *parting)
{

}

static void
snd_user_change (htlc_t *to, htlc_t *htlc)
{

}

static void
snd_error (htlc_t *htlc, const u_int8_t *str, u_int16_t len)
{
//	cp_snd_linef(htlc, "NOTICE %s: %s", htlc->name, str);
}

struct protocol_function irc_functions[] = {
	{ RCV_MAGIC, { rcv_magic } },
	{ SND_CHAT, { snd_chat } },
	{ SND_MSG, { snd_msg } },
	{ SND_ERROR, { snd_error } },
	{ SND_USER_PART, { snd_user_part } },
	{ SND_USER_CHANGE, { snd_user_change } }
};
#define nirc_functions (sizeof(irc_functions)/sizeof(struct protocol_function))

#ifdef MODULE
#define irc_protocol_init module_init
#endif

void
irc_protocol_init (void)
{
	irc_proto = protocol_register("irc");
	if (!irc_proto)
		return;
	protocol_register_functions(irc_proto, irc_functions, nirc_functions);
}
